@@ -7,12 +7,17 @@ import (
 	"context"
 	"sort"
 	"strings"
+	"time"
 
 	"github.com/go-kit/kit/log"
 	"github.com/opentracing/opentracing-go"
 	"github.com/pkg/errors"
+	"github.com/prometheus/client_golang/prometheus"
 	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/storage"
+
+	"github.com/thanos-io/thanos/pkg/gate"
+	"github.com/thanos-io/thanos/pkg/store"
 	"github.com/thanos-io/thanos/pkg/store/storepb"
 	"github.com/thanos-io/thanos/pkg/tracing"
 )
@@ -22,50 +27,66 @@ import (
 // When the replicaLabels argument is not empty it overwrites the global replicaLabels flag. This allows specifying
 // replicaLabels at query time.
 // maxResolutionMillis controls downsampling resolution that is allowed (specified in milliseconds).
-// partialResponse controls `partialResponseDisabled` option of StoreAPI and partial response behaviour of proxy.
-type QueryableCreator func(deduplicate bool, replicaLabels []string, maxResolutionMillis int64, partialResponse, skipChunks bool) storage.Queryable
+// partialResponse controls `partialResponseDisabled` option of StoreAPI and partial response behavior of proxy.
+type QueryableCreator func(deduplicate bool, replicaLabels []string, storeMatchers [][]storepb.LabelMatcher, maxResolutionMillis int64, partialResponse, skipChunks bool) storage.Queryable
 
 // NewQueryableCreator creates QueryableCreator.
-func NewQueryableCreator(logger log.Logger, proxy storepb.StoreServer) QueryableCreator {
-	return func(deduplicate bool, replicaLabels []string, maxResolutionMillis int64, partialResponse, skipChunks bool) storage.Queryable {
+func NewQueryableCreator(logger log.Logger, reg prometheus.Registerer, proxy storepb.StoreServer, maxConcurrentSelects int, selectTimeout time.Duration) QueryableCreator {
+	keeper := gate.NewKeeper(reg)
+
+	return func(deduplicate bool, replicaLabels []string, storeMatchers [][]storepb.LabelMatcher, maxResolutionMillis int64, partialResponse, skipChunks bool) storage.Queryable {
 		return &queryable{
-			logger:              logger,
-			replicaLabels:       replicaLabels,
-			proxy:               proxy,
-			deduplicate:         deduplicate,
-			maxResolutionMillis: maxResolutionMillis,
-			partialResponse:     partialResponse,
-			skipChunks:          skipChunks,
+			logger:               logger,
+			reg:                  reg,
+			replicaLabels:        replicaLabels,
+			storeMatchers:        storeMatchers,
+			proxy:                proxy,
+			deduplicate:          deduplicate,
+			maxResolutionMillis:  maxResolutionMillis,
+			partialResponse:      partialResponse,
+			skipChunks:           skipChunks,
+			gateKeeper:           keeper,
+			maxConcurrentSelects: maxConcurrentSelects,
+			selectTimeout:        selectTimeout,
 		}
 	}
 }
 
 type queryable struct {
-	logger              log.Logger
-	replicaLabels       []string
-	proxy               storepb.StoreServer
-	deduplicate         bool
-	maxResolutionMillis int64
-	partialResponse     bool
-	skipChunks          bool
+	logger               log.Logger
+	reg                  prometheus.Registerer
+	replicaLabels        []string
+	storeMatchers        [][]storepb.LabelMatcher
+	proxy                storepb.StoreServer
+	deduplicate          bool
+	maxResolutionMillis  int64
+	partialResponse      bool
+	skipChunks           bool
+	gateKeeper           *gate.Keeper
+	maxConcurrentSelects int
+	selectTimeout        time.Duration
 }
 
 // Querier returns a new storage querier against the underlying proxy store API.
 func (q *queryable) Querier(ctx context.Context, mint, maxt int64) (storage.Querier, error) {
-	return newQuerier(ctx, q.logger, mint, maxt, q.replicaLabels, q.proxy, q.deduplicate, int64(q.maxResolutionMillis), q.partialResponse, q.skipChunks), nil
+	return newQuerier(ctx, q.logger, q.reg, mint, maxt, q.replicaLabels, q.storeMatchers, q.proxy, q.deduplicate, q.maxResolutionMillis, q.partialResponse, q.skipChunks, q.gateKeeper.NewGate(q.maxConcurrentSelects), q.selectTimeout), nil
 }
 
 type querier struct {
 	ctx                 context.Context
 	logger              log.Logger
+	reg                 prometheus.Registerer
 	cancel              func()
 	mint, maxt          int64
 	replicaLabels       map[string]struct{}
+	storeMatchers       [][]storepb.LabelMatcher
 	proxy               storepb.StoreServer
 	deduplicate         bool
 	maxResolutionMillis int64
 	partialResponse     bool
 	skipChunks          bool
+	selectGate          gate.Gate
+	selectTimeout       time.Duration
 }
 
 // newQuerier creates implementation of storage.Querier that fetches data from the proxy
@@ -73,13 +94,16 @@ type querier struct {
 func newQuerier(
 	ctx context.Context,
 	logger log.Logger,
+	reg prometheus.Registerer,
 	mint, maxt int64,
 	replicaLabels []string,
+	storeMatchers [][]storepb.LabelMatcher,
 	proxy storepb.StoreServer,
 	deduplicate bool,
 	maxResolutionMillis int64,
-	partialResponse bool,
-	skipChunks bool,
+	partialResponse, skipChunks bool,
+	selectGate gate.Gate,
+	selectTimeout time.Duration,
 ) *querier {
 	if logger == nil {
 		logger = log.NewNopLogger()
@@ -91,12 +115,17 @@ func newQuerier(
 		rl[replicaLabel] = struct{}{}
 	}
 	return &querier{
-		ctx:                 ctx,
-		logger:              logger,
-		cancel:              cancel,
+		ctx:           ctx,
+		logger:        logger,
+		reg:           reg,
+		cancel:        cancel,
+		selectGate:    selectGate,
+		selectTimeout: selectTimeout,
+
 		mint:                mint,
 		maxt:                maxt,
 		replicaLabels:       rl,
+		storeMatchers:       storeMatchers,
 		proxy:               proxy,
 		deduplicate:         deduplicate,
 		maxResolutionMillis: maxResolutionMillis,
@@ -124,10 +153,12 @@ func (s *seriesServer) Send(r *storepb.SeriesResponse) error {
 		return nil
 	}
 
-	if r.GetSeries() == nil {
-		return errors.New("no seriesSet")
+	if r.GetSeries() != nil {
+		s.seriesSet = append(s.seriesSet, *r.GetSeries())
+		return nil
 	}
-	s.seriesSet = append(s.seriesSet, *r.GetSeries())
+
+	// Unsupported field, skip.
 	return nil
 }
 
@@ -135,43 +166,32 @@ func (s *seriesServer) Context() context.Context {
 	return s.ctx
 }
 
-type resAggr int
-
-const (
-	resAggrAvg resAggr = iota
-	resAggrCount
-	resAggrSum
-	resAggrMin
-	resAggrMax
-	resAggrCounter
-)
-
 // aggrsFromFunc infers aggregates of the underlying data based on the wrapping
 // function of a series selection.
-func aggrsFromFunc(f string) ([]storepb.Aggr, resAggr) {
+func aggrsFromFunc(f string) []storepb.Aggr {
 	if f == "min" || strings.HasPrefix(f, "min_") {
-		return []storepb.Aggr{storepb.Aggr_MIN}, resAggrMin
+		return []storepb.Aggr{storepb.Aggr_MIN}
 	}
 	if f == "max" || strings.HasPrefix(f, "max_") {
-		return []storepb.Aggr{storepb.Aggr_MAX}, resAggrMax
+		return []storepb.Aggr{storepb.Aggr_MAX}
 	}
 	if f == "count" || strings.HasPrefix(f, "count_") {
-		return []storepb.Aggr{storepb.Aggr_COUNT}, resAggrCount
+		return []storepb.Aggr{storepb.Aggr_COUNT}
 	}
 	// f == "sum" falls through here since we want the actual samples.
 	if strings.HasPrefix(f, "sum_") {
-		return []storepb.Aggr{storepb.Aggr_SUM}, resAggrSum
+		return []storepb.Aggr{storepb.Aggr_SUM}
 	}
-	if f == "increase" || f == "rate" {
-		return []storepb.Aggr{storepb.Aggr_COUNTER}, resAggrCounter
+	if f == "increase" || f == "rate" || f == "irate" || f == "resets" {
+		return []storepb.Aggr{storepb.Aggr_COUNTER}
 	}
 	// In the default case, we retrieve count and sum to compute an average.
-	return []storepb.Aggr{storepb.Aggr_COUNT, storepb.Aggr_SUM}, resAggrAvg
+	return []storepb.Aggr{storepb.Aggr_COUNT, storepb.Aggr_SUM}
 }
 
-func (q *querier) Select(params *storage.SelectParams, ms ...*labels.Matcher) (storage.SeriesSet, storage.Warnings, error) {
-	if params == nil {
-		params = &storage.SelectParams{
+func (q *querier) Select(_ bool, hints *storage.SelectHints, ms ...*labels.Matcher) storage.SeriesSet {
+	if hints == nil {
+		hints = &storage.SelectHints{
 			Start: q.mint,
 			End:   q.maxt,
 		}
@@ -181,31 +201,78 @@ func (q *querier) Select(params *storage.SelectParams, ms ...*labels.Matcher) (s
 	for i, m := range ms {
 		matchers[i] = m.String()
 	}
-	span, ctx := tracing.StartSpan(q.ctx, "querier_select", opentracing.Tags{
-		"minTime":  params.Start,
-		"maxTime":  params.End,
+
+	// The querier has a context but it gets canceled, as soon as query evaluation is completed, by the engine.
+	// We want to prevent this from happening for the async storea API calls we make while preserving tracing context.
+	ctx := tracing.CopyTraceContext(context.Background(), q.ctx)
+	ctx, cancel := context.WithTimeout(ctx, q.selectTimeout)
+	span, ctx := tracing.StartSpan(ctx, "querier_select", opentracing.Tags{
+		"minTime":  hints.Start,
+		"maxTime":  hints.End,
 		"matchers": "{" + strings.Join(matchers, ",") + "}",
 	})
-	defer span.Finish()
 
-	sms, err := translateMatchers(ms...)
+	promise := make(chan storage.SeriesSet, 1)
+	go func() {
+		defer close(promise)
+
+		var err error
+		tracing.DoInSpan(ctx, "querier_select_gate_ismyturn", func(ctx context.Context) {
+			err = q.selectGate.Start(ctx)
+		})
+		if err != nil {
+			promise <- storage.ErrSeriesSet(errors.Wrap(err, "failed to wait for turn"))
+			return
+		}
+		defer q.selectGate.Done()
+
+		span, ctx := tracing.StartSpan(ctx, "querier_select_select_fn")
+		defer span.Finish()
+
+		set, err := q.selectFn(ctx, hints, ms...)
+		if err != nil {
+			promise <- storage.ErrSeriesSet(err)
+			return
+		}
+
+		promise <- set
+	}()
+
+	return &lazySeriesSet{create: func() (storage.SeriesSet, bool) {
+		defer cancel()
+		defer span.Finish()
+
+		// Only gets called once, for the first Next() call of the series set.
+		set, ok := <-promise
+		if !ok {
+			return storage.ErrSeriesSet(errors.New("channel closed before a value received")), false
+		}
+		return set, set.Next()
+	}}
+}
+
+func (q *querier) selectFn(ctx context.Context, hints *storage.SelectHints, ms ...*labels.Matcher) (storage.SeriesSet, error) {
+	sms, err := storepb.TranslatePromMatchers(ms...)
 	if err != nil {
-		return nil, nil, errors.Wrap(err, "convert matchers")
+		return nil, errors.Wrap(err, "convert matchers")
 	}
 
-	queryAggrs, resAggr := aggrsFromFunc(params.Func)
+	aggrs := aggrsFromFunc(hints.Func)
+
+	// TODO: Pass it using the SerieRequest instead of relying on context
+	ctx = context.WithValue(ctx, store.StoreMatcherKey, q.storeMatchers)
 
 	resp := &seriesServer{ctx: ctx}
 	if err := q.proxy.Series(&storepb.SeriesRequest{
-		MinTime:                 params.Start,
-		MaxTime:                 params.End,
+		MinTime:                 hints.Start,
+		MaxTime:                 hints.End,
 		Matchers:                sms,
 		MaxResolutionWindow:     q.maxResolutionMillis,
-		Aggregates:              queryAggrs,
+		Aggregates:              aggrs,
 		PartialResponseDisabled: !q.partialResponse,
 		SkipChunks:              q.skipChunks,
 	}, resp); err != nil {
-		return nil, nil, errors.Wrap(err, "proxy Series()")
+		return nil, errors.Wrap(err, "proxy Series()")
 	}
 
 	var warns storage.Warnings
@@ -216,28 +283,27 @@ func (q *querier) Select(params *storage.SelectParams, ms ...*labels.Matcher) (s
 	if !q.isDedupEnabled() {
 		// Return data without any deduplication.
 		return &promSeriesSet{
-			mint: q.mint,
-			maxt: q.maxt,
-			set:  newStoreSeriesSet(resp.seriesSet),
-			aggr: resAggr,
-		}, warns, nil
+			mint:  q.mint,
+			maxt:  q.maxt,
+			set:   newStoreSeriesSet(resp.seriesSet),
+			aggrs: aggrs,
+			warns: warns,
+		}, nil
 	}
 
-	// TODO(fabxc): this could potentially pushed further down into the store API
-	// to make true streaming possible.
+	// TODO(fabxc): this could potentially pushed further down into the store API to make true streaming possible.
 	sortDedupLabels(resp.seriesSet, q.replicaLabels)
-
 	set := &promSeriesSet{
-		mint: q.mint,
-		maxt: q.maxt,
-		set:  newStoreSeriesSet(resp.seriesSet),
-		aggr: resAggr,
+		mint:  q.mint,
+		maxt:  q.maxt,
+		set:   newStoreSeriesSet(resp.seriesSet),
+		aggrs: aggrs,
+		warns: warns,
 	}
 
-	// The merged series set assembles all potentially-overlapping time ranges
-	// of the same series into a single one. The series are ordered so that equal series
-	// from different replicas are sequential. We can now deduplicate those.
-	return newDedupSeriesSet(set, q.replicaLabels), warns, nil
+	// The merged series set assembles all potentially-overlapping time ranges of the same series into a single one.
+	// TODO(bwplotka): We could potentially dedup on chunk level, use chunk iterator for that when available.
+	return newDedupSeriesSet(set, q.replicaLabels, len(aggrs) == 1 && aggrs[0] == storepb.Aggr_COUNTER), nil
 }
 
 // sortDedupLabels re-sorts the set so that the same series with different replica
@@ -267,7 +333,12 @@ func (q *querier) LabelValues(name string) ([]string, storage.Warnings, error) {
 	span, ctx := tracing.StartSpan(q.ctx, "querier_label_values")
 	defer span.Finish()
 
-	resp, err := q.proxy.LabelValues(ctx, &storepb.LabelValuesRequest{Label: name, PartialResponseDisabled: !q.partialResponse})
+	resp, err := q.proxy.LabelValues(ctx, &storepb.LabelValuesRequest{
+		Label:                   name,
+		PartialResponseDisabled: !q.partialResponse,
+		Start:                   q.mint,
+		End:                     q.maxt,
+	})
 	if err != nil {
 		return nil, nil, errors.Wrap(err, "proxy LabelValues()")
 	}
@@ -285,7 +356,11 @@ func (q *querier) LabelNames() ([]string, storage.Warnings, error) {
 	span, ctx := tracing.StartSpan(q.ctx, "querier_label_names")
 	defer span.Finish()
 
-	resp, err := q.proxy.LabelNames(ctx, &storepb.LabelNamesRequest{PartialResponseDisabled: !q.partialResponse})
+	resp, err := q.proxy.LabelNames(ctx, &storepb.LabelNamesRequest{
+		PartialResponseDisabled: !q.partialResponse,
+		Start:                   q.mint,
+		End:                     q.maxt,
+	})
 	if err != nil {
 		return nil, nil, errors.Wrap(err, "proxy LabelNames()")
 	}
